#!/usr/bin/perl
#
# $Id$
#
# (C) ITEG IT-Engineers GmbH 2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

use strict;


sub initCFTFunctions() { 

  $main::version = "0.4.1-6";
  if (length($main::scriptName) > 0) {
    $main::versionMessage = $main::scriptName." (config-file-tools ".$main::version.")\n";
  } else {
    $main::versionMessage = "config-file-tools ".$main::version."\n";
  }

  $main::format = "n=v";
  $main::verbosity = 1;
  $main::mode = "u";
  @main::inputFiles = ();
  %main::optionsToApply = ();
  %main::optionsCommentsToApply = ();
  @main::targetFiles = ();
  $main::commentForChanges = "";
  $main::documentFormerValue = 0;
  $main::timestampChanges = 0;

  $main::errors = "";
  $main::warnings = "";

  my $argc = scalar(@ARGV);
  my $i = 0;
  while ($i < $argc) { 
    my $curOpt = $ARGV[$i];
    #print "DEBUG: Investigating arg ".$i.", '".$curOpt."'\n";
    $i++;
    if ($curOpt eq "-a") { 
      $main::mode = "a";
      next;
    }
    if ($curOpt eq "-c") {
      $main::commentForChanges = $ARGV[$i];
      $i++;
      next;
    }
    if ($curOpt eq "-d") { 
      $main::verbosity += 2;
      next;
    }
    if ($curOpt eq "--help") {
      print $main::helpMessage;
      exit 0;
    }
    if ($curOpt eq "-i") {
      parseConfFileToApply($ARGV[$i]);
      $i++;
      next;
    }
    if ($curOpt eq "-k") { 
      $main::documentFormerValue = 1;
      next;
    }
    if ($curOpt eq "-n") { 
      $main::mode = "n";
      next;
    }
    if ($curOpt eq "-q") { 
      $main::verbosity = 0;
      next;
    }
    if ($curOpt eq "-s") {
      parseOptionToApply($ARGV[$i]);
      $i++;
      next;
    }
    if ($curOpt eq "-t") { 
      $main::timestampChanges = 1;
      next;
    }
    if ($curOpt eq "-u") { 
      $main::mode = "u";
      next;
    }
    if ($curOpt eq "--version") {
      print $main::versionMessage;
      exit 0;
    }
    if ($curOpt eq "-v") {
      $main::verbosity += 1;
      next;
    }
    if ( -f $curOpt ) {
      #print "DEBUG: Using as target file: ".$curOpt."\n";
      push @main::targetFiles, $curOpt;
    } else {
      $main::warnings .= "WARNING: No such file: ".$curOpt."\n";
    }
  }

  $main::nrTargetFiles = scalar(@main::targetFiles);
  if ($main::nrTargetFiles == 0) {
    $main::errors .= "FATAL: No existing target files specified.\n";
  }

  $main::nrOptionsToApply = scalar(keys(%main::optionsToApply));
  if ($main::nrOptionsToApply == 0) {
    $main::errors .= "FATAL: No options to apply available.\n";
  }

  if (length($main::errors) > 0) {
    cleanupAndDie($main::errors);
  }
  if (length($main::warnings) > 0) {
    print $main::warnings;
    print "Continuing anyway.\n";
  }

  if ($main::verbosity gt 2) {
    if ($main::mode eq "a") {
      print "Mode: (a)ppend name=value if no name= line there yet.\n";
    } else {
      if ($main::mode eq "n") {
        print "Mode: (n)ew values only, i.e. append-only.\n";
      } else {
        print "Mode: (u)pdate only existing name=value lines.\n";
      }
    }
    print "Options to apply:\n";
    foreach my $confName (sort(keys(%main::optionsToApply))) {
      print "  '".$confName."' to '".$main::optionsToApply{$confName}."'\n";
    }
    print "Target files:\n";
    foreach my $targetFile (sort(@main::targetFiles)) {
      print "  ".$targetFile."\n";
    }
    print "Comment for changes: '".$main::commentForChanges."'\n";
    print "Timestamp changes: '".$main::timestampChanges."'\n";
    print "Document former value: '".$main::documentFormerValue."'\n";
  }

  $main::nrFilesChanged = 0;
}


# functions


# cleanup

sub cleanup() {
  #if (scalar(@main::filesToCleanup) > 0) {
  #  if ($main::verbosity > 1) { print "cleanup(): rm'ing these files:\n".join("\n",@main::filesToCleanup)."\n"; }
  #  unlink @main::filesToCleanup;
  #}
  #if (scalar(@main::dirsToCleanup) > 0) {
  #  if ($main::verbosity > 1) { print "cleanup(): rm'ing these directories:\n".join("\n",@main::dirsToCleanup)."\n"; }
  #  system "/bin/rm -rf ".join(" ", @main::dirsToCleanup);
  #}
}


sub cleanupAndDie( $ ) {
  # $1 = message
  my $message = shift;
  cleanup();
  if ($message =~ /\n$/)
  { $message .= "\n"; }
  if ($message =~ /\n$/)
  { $message .= "\n"; }
  die $message;
}


# file helpers


sub printToFile( $ $ ; $ ) {
  # $1 = filename
  # $2 = text
  # $3 = mode, optional, defaults to 0755
  my $filename = shift;
  my $text = shift;
  my $mode = shift;
  if ($mode eq "") { $mode = 0755; }
  open (TMPFILE, ">".$filename) || cleanupAndDie("FATAL: Cannot write ".$filename."\n");
  print TMPFILE $text;
  close TMPFILE;
  if ($filename =~ /\.(sh|pl)$/) {
    chmod $mode, $filename || cleanupAndDie("FATAL: Cannot chmod ".$filename."\n");
  }
}


sub readCompleteFile( $ ) {
  # $1 = name of text file
  # returns content of file
  my $filename=shift;
  my $text;
  my $defaultLF=$/;
  undef $/;
  open (TMPFILE, "<".$filename) || cleanupAndDie("FATAL: Cannot read ".$filename);
  $text=<TMPFILE>;
  close TMPFILE;
  $/=$defaultLF;
  return $text;
}


# parsing


sub parseConfFileToApply( $ ) {
  my $confFile = shift;
  open (CONFFILE, "<".$confFile)  || cleanupAndDie("FATAL: Cannot read $confFile.\n");
  my $lineNr = 0;
  my $commentParagraph = "";
  while ( my $curLine = <CONFFILE> ) {
    chomp $curLine;
    $lineNr++;
    if ( $curLine =~ m/^\s*[#;]/ ) {
      $commentParagraph .= $curLine."\n";
      next;
    }
    if ( $curLine =~ m/^\s*$/ ) { 
      $commentParagraph = "";
      next;
    }
    if ( $curLine =~ m/^\s*\[/ ) { 
      #if ( $curLine =~ m/^\[.*\]\s*$/ ) { 
      # ini style section title, for now just treat like an empty line
      $commentParagraph = "";
      next;
    }
    my ( $confName, $confValue ) = split / *= */, $curLine, 2;
    if (length($confName) == 0) {
      $main::warnings .= "Ignoring invalid line ".$confFile.":".$lineNr."\n";
      $commentParagraph = "";
      next;
    }
    if (length($confValue) == 0) {
      $main::warnings .= "Using empty value for ".$confName." from ".$confFile.":".$lineNr."\n";
    }
    $main::optionsToApply{$confName} = $confValue;
    $main::optionsCommentsToApply{$confName} = $commentParagraph;
  }
 
}


sub parseOptionToApply( $ ) {
  my $curLine = shift;
  #next if ( $curLine =~ m/^\s*[#;]/ );
  #next if ( $curLine =~ m/^\s*$/ );
  my ( $confName, $confValue ) = split / *= */, $curLine, 2;
  if (length($confName) == 0) {
    $main::warnings .= "Ignoring invalid line ".$curLine."\n";
    return;
  }
  if (length($confValue) == 0) {
    $main::warnings .= "Using empty value for ".$confName."\n";
  }
  $main::optionsToApply{$confName} = $confValue;
}


sub updateConfFile( $ ) {
  my $confFile = shift;
  my $newConf = "";
  my $nrChanges = 0;
  my %existingConfOptions = ();
  my ($sec, $min, $hour, $day, $mon, $year) = localtime();
  my $timeFormated = sprintf("%04d-%02d-%02d %02d:%02d:%02d", 1900+$year, $mon + 1, $day, $hour, $min, $sec);
  open (CONFFILE, "<".$confFile)  || cleanupAndDie("FATAL: Cannot read $confFile.\n");
  my $lineNr = 0;
  while ( my $curLine = <CONFFILE> ) {
    chomp $curLine;
    $lineNr++;
    if ( $curLine =~ m/^\s*[#;]/ ) { 
      if ($main::verbosity gt 3) { print "Ignoring comment line ".$confFile.":".$lineNr."\n"; }
      $newConf .= $curLine."\n";
      next;
    }
    if ( $curLine =~ m/^\s*$/ ) {
      if ($main::verbosity gt 3) { print "Ignoring empty line ".$confFile.":".$lineNr."\n"; }
      $newConf .= $curLine."\n";
      next;
    }
    if ( $curLine =~ m/^\s\[/ ) {
      if ($main::verbosity gt 3) { print "Ignoring INI section title line ".$confFile.":".$lineNr."\n"; }
      $newConf .= $curLine."\n";
      next;
    }
    my ( $confName, $confValue ) = split / *= */, $curLine, 2;
    if (length($confName) == 0) {
      if ($main::verbosity gt 0) { print "WARNING: Ignoring invalid line ".$confFile.":".$lineNr."\n"; }
      $newConf .= $curLine."\n";
      next;
    }
    $existingConfOptions{$confName} = $confValue;
    if ($main::mode eq "n") {
      if ($main::verbosity gt 2) { print "In (n)ew-only mode ignoring current value of option '".$confName."' in ".$confFile.":".$lineNr."\n"; }
      $newConf .= $curLine."\n";
      next;
    }
    if (! defined $main::optionsToApply{$confName}) {
      if ($main::verbosity gt 2) { print "Not interested in option '".$confName."' in ".$confFile.":".$lineNr."\n"; }
      $newConf .= $curLine."\n";
      next;
    }
    if ($confValue eq $main::optionsToApply{$confName}) {
      if ($main::verbosity gt 2) { print "Value not changing for '".$confName."' in ".$confFile.":".$lineNr.", '".$confFile."'='".$main::optionsToApply{$confName}."'\n"; }
      $newConf .= $curLine."\n";
      next;
    }
    if ($main::verbosity gt 1) { print "Value changing for '".$confName."' in ".$confFile.":".$lineNr.", '".$confValue."'=>'".$main::optionsToApply{$confName}."'\n"; }
    if ($main::documentFormerValue gt 0) {
      $newConf .= "#".$curLine."\n";
    }
    if ($main::commentForChanges gt 0) {
      if ($main::timestampChanges gt 0) {
        $newConf .= "# ".$main::commentForChanges.", changed at ".$timeFormated." to:\n";
      } else {
        $newConf .= "# ".$main::commentForChanges."\n";
      }
    } else {
      if ($main::timestampChanges gt 0) {
        $newConf .= "# changed at ".$timeFormated." to:\n";
      }
    }
    $newConf .= $confName."=".$main::optionsToApply{$confName}."\n";
    $nrChanges++;
  }
  if (($main::mode eq "a") or ($main::mode eq "n")) {
    # make sure we have a clear line
    $newConf .= "\n";
    for my $confName (sort(keys(%main::optionsToApply))) {
      if (! defined($existingConfOptions{$confName})) {
        $newConf .= "\n";
        if ($main::verbosity gt 1) { print "Appending mssing value to ".$confFile.": '".$confName."'='".$main::optionsToApply{$confName}."'\n"; }
        if ($main::commentForChanges gt 0) {
          if ($main::timestampChanges gt 0) {
            $newConf .= "# ".$main::commentForChanges.", added at ".$timeFormated."\n";
          } else {
            $newConf .= "# ".$main::commentForChanges."\n";
          }
        } else {
          if ($main::timestampChanges gt 0) {
            $newConf .= "# added at ".$timeFormated."\n";
          }
        }
        if (defined $main::optionsCommentsToApply{$confName}) {
          $newConf .= $main::optionsCommentsToApply{$confName};
        }
        $newConf .= $confName."=".$main::optionsToApply{$confName}."\n";
        $nrChanges++;
      }
    }
  }
  if ($nrChanges == 0) {
    if ($main::verbosity gt 1) { print "No changes in ".$confFile."\n"; }
    return;
  }
  if ($main::verbosity gt 0) { print "Writing new ".$confFile."\n"; }
  printToFile($confFile, $newConf);
  $main::nrFilesChanged++;
}


# EOF
1;

